#!/usr/bin/env python

import sys
import os

from twisted.internet import reactor, stdio, task
from twisted.python import log
from openid.store.filestore import FileOpenIDStore
import itsdangerous
from llarma import GamestateTracker
from clize import clize, run

from aagidopenidgateway.misc import TokenStorage, Janitor
from aagidopenidgateway.web import make_site
from aagidopenidgateway.arma import ArmaTokenSender

@clize(force_positional=('listen', 'port'))
def main(instance_directory, listen='localhost', port=8443,
         address=None,
         arma_host=None, arma_port=4534,
         token_expiry=10 * 60, idle_time=10 * 60,
         debug=False
         ):
    """When run as an armagetron advanced script, hosts a webserver that
    handles incoming openid auths.

    instance_directory: The path to a directory where to find secret_key and
    where openid associations will be stored.

    listen: The address the http servers listens on

    port: The port the http server listens on

    address: The hostname to expect requests on, if different
    from listen:port

    arma_host: The hostname part of the address to the armagetron server.
    Defaults to the value of listen

    arma_port: The port part of the address to the armagetron server.

    token_expiry: How long to wait for a user to use the token given
    in-game

    idle_time: How long to wait before kicking inactive users in the server

    debug: Set the app into (unsafe) debug mode.

    Times are in seconds.
    """
    log.startLogging(sys.stderr)

    if not address:
        if port != 80:
            address = '{0}:{1}'.format(listen, port)
        else:
            address = listen

    if not arma_host:
        arma_host = listen

    secret_key_file = os.path.join(instance_directory, 'secret_key')
    with open(secret_key_file, 'rb') as f:
        if os.fstat(f.fileno()).st_mode & 0o077:
            log.msg("'{0}' must not be readable by other users, aborting."
                  .format(secret_key_file))
            sys.exit(2)
        secret_key = f.read()
    oidstore = FileOpenIDStore(os.path.join(instance_directory, 'oid'))

    storage = TokenStorage(token_expiry)

    if debug:
        serializer = itsdangerous.Serializer(secret_key)
    else:
        serializer = itsdangerous.JSONWebSignatureSerializer(secret_key)

    site = make_site(oidstore, storage,
                     arma_host, arma_port,
                     address, serializer)
    site.displayTracebacks = debug
    reactor.listenTCP(port, site, interface=listen)

    tokensender = ArmaTokenSender(storage, address, idle_time)
    stdio.StandardIO(GamestateTracker(tokensender))

    task.LoopingCall(Janitor(storage, tokensender)).start(60)

    reactor.run()

def entry_point():
    run(main)

if __name__ == '__main__':
    entry_point()
