from time import time
from functools import partial

from twisted.internet import reactor
from twisted.python import log

from llarma import LadderlogProtocol


def eachline(func, text):
    return '\n'.join(func(line) for line in text.split('\n'))

def colorize_func(color):
    return partial(eachline, lambda s: color + s)

info_color = colorize_func('0xffff7f')

# arma interaction

class ArmaTokenSender(LadderlogProtocol):
    """"""
    def __init__(self, storage, hostname, idle_time):
        self.storage = storage
        self.hostname = hostname
        self.idle_time = idle_time

        self.storage.subscribe(self.on_use)

        # player -> expiry map
        self.users = {}
        # authname -> player map
        self.auth_to_players = {}

    def player_message(self, player, message):
        self.transport.command('PLAYER_MESSAGE', player, message)

    def info(self, player, message):
        self.player_message(player, info_color(message))

    def ladderlog_player_entered(self, player, ip, nick):
        self.users[player] = time() + self.idle_time
        self.info(player, 'Please authenticate to proceed.')

    def ladderlog_player_renamed(self, oldusername, player, ip, nick):
        if oldusername == player.logname:
            return
        if player.authed:
            self.auth_to_players[player.logname] = player
            self.give_token(player)

    def ladderlog_player_left(self, player, ip):
        self.users.pop(player, None)
        delete = []
        for authname, player_ in self.auth_to_players.items():
            if player_ == player:
                delete.append(authname)
        for key in delete:
            del self.auth_to_players[key]

    def on_use(self, authname):
        try:
            player = self.auth_to_players[authname]
        except KeyError:
            pass
        else:
            self.transport.command('KICK', player, 'Authentication succesful.')

    def format_token(self, token):
        return 'http://{0}/auth/{1}'.format(
            self.hostname,
            '-'.join(token[i:i+4] for i in range(0, len(token), 4))
            )

    def give_token(self, player):
        token = log.callWithLogger(self.storage,
            self.storage.new_token,
            player.logname, player.ip)
        self.info(player,
            "Navigate to this address to continue:\n"
            "    0xff7f7f{0}"
            .format(self.format_token(token))
        )

    def cleanup(self):
        now = time()
        to_delete = []
        for user, expiry in self.users.iteritems():
            if now > expiry:
                to_delete.append(user)
        for user in to_delete:
            self.transport.command('KICK', user, "Idle time exceeded.")

    def connectionLost(self, reason):
        # No point going on if stdin was closed
        if reactor.running:
            reactor.stop()
