from random import SystemRandom
from time import time

from twisted.python import log

ALPHABET = 'abcdefghjkmnprtuvwxyz2346789'
TOKEN_LEN = 12
class TokenStorage(object):
    """Creates and checks auth tokens"""
    def __init__(self, token_expiry):
        """Initializer

        request_expiry: how long to wait before forgetting an oid auth req

        token_expiry: how long to wait before invalidating a token

        """
        self.token_expiry = token_expiry

        # store the previously generated token so we can know
        # if we're running on a dilbert number generator
        self.last_token = None
        self.random = SystemRandom()

        # id -> request, expiry pairs
        self.requests = {}
        # token -> authname, ip, oid_request, expiry pairs
        self.tokens = {}

        # callbacks
        self.on_use = []

    def logPrefix(self):
        return type(self).__name__

    def requests_from_session(self, request):
        sess = request.session
        ret = {}
        if 'authreqs' not in sess:
            return ret
        expired = []
        for ident in sess['authreqs']:
            try:
                ret[ident] = self.requests[ident]
            except KeyError:
                expired.append(ident)
        if expired:
            sess['authreqs'] = [
                ident
                for ident in sess['authreqs']
                if ident not in expired]
            request.push_session()
        return ret

    def _random_token(self):
        return ''.join(self.random.choice(ALPHABET)
                       for i in range(TOKEN_LEN))

    def _new_token(self):
        for i in range(5):
            token = self._random_token()
            if token not in self.tokens and token != self.last_token:
                self.last_token = token
                return token
        else:
            raise RuntimeError("Couldn't get a random token")

    def new_token(self, authname, ip):
        token = self._new_token()
        expiry = time() + self.token_expiry
        self.tokens[token] = authname, ip, expiry
        log.msg('Created a token for {0} from {1}'.format(authname, ip))
        return token

    def use_token(self, ip, token):
        if token in self.tokens:
            authname, ip_, expiry = self.tokens[token]
            if ip == ip_ and expiry > time():
                del self.tokens[token]
                log.msg('Used token for {0!r} from {1} for {2!r}'
                    .format(authname, ip, None))
                self.run_callbacks(self.on_use, authname)
                return authname
        return None

    def subscribe(self, on_use):
        if on_use:
            self.on_use.append(on_use)

    def run_callbacks(self, source, *args):
        for func in source:
            log.callWithLogger(self, func, *args)

    def cleanup(self):
        now = time()

        for token, info in list(self.tokens.items()):
            authname, ip, expiry = info
            if now > expiry:
                del self.tokens[authname]
                log.msg('Expiring token for {0} from {1}'.format(authname, ip))

class Janitor(object):
    """when an instance is called, calls obj.cleanup() on the objects
    passed to the constructor"""

    def __init__(self, *to_clean):
        self.to_clean = to_clean

    def cleanup(self):
        for obj in self.to_clean:
            obj.cleanup()

    __call__ = cleanup

