from time import time

from pkg_resources import resource_filename
from openid.server import server as openidserver
from openid import message as oidmessage

from twisted.web.resource import Resource, NoResource
from twisted.web.server import Site
from twisted.web.static import File
from twisted.python import log

from aagidopenidgateway.webf import SessionGetter, NiceRequest

def gid_from_oid(oid):
    gid = oid[8:].partition('/')[2].strip('/')
    if not gid:
        raise ValueError('no gid in: ' + oid)
    return gid

def oid_from_gid(request, gid):
    return 'http://{0}/{1}'.format(request.getHeader('host'), gid)

def serialize_check_id(authreq):
    return authreq.message.toPostArgs(), authreq.op_endpoint
    return (
        authreq.identity, authreq.return_to, authreq.trust_root,
        authreq.immediate, authreq.assoc_handle, authreq.op_endpoint,
        authreq.claimed_id
        )

def unserialize_check_id(serialized):
    return openidserver.CheckIDRequest.fromMessage(
        oidmessage.Message.fromPostArgs(serialized[0]), serialized[1])

def expire_check_ids(serialized):
    oldest_time = time() - REQUEST_EXPIRY
    for key, (time_created, value) in serialized.items():
        if time_created > oldest_time:
            yield key, (time_created, value)

REQUEST_EXPIRY = 60 * 30
def unserialize_check_ids(serialized):
    for key, (time_created, value) in serialized.items():
        yield key, unserialize_check_id(value)

def authreqs_from_session(sess):
    valid = dict(expire_check_ids(sess.get('authreqs', {})))
    sess['authreqs'] = valid
    return dict(unserialize_check_ids(valid))

class OpenIdEndpoint(Resource):
    isLeaf = True

    def __init__(self, openidserver, storage):
        Resource.__init__(self)
        self.oidserver = openidserver
        self.storage = storage

    def start_auth(self, request, oid_request, error=None):
        sess = request.session
        authreqs = dict(expire_check_ids(sess.get('authreqs', {})))
        authreqs[oid_request.trust_root] = (
            time(), serialize_check_id(oid_request))
        sess['authreqs'] = authreqs
        request.push_session()
        return request.redirect_page('/user?' if 'user' in sess else '/auth')

    def oid_response(self, request, oid_response):
        encoded = self.oidserver.encodeResponse(oid_response)
        request.setResponseCode(encoded.code)
        for header, val in encoded.headers.iteritems():
            request.setHeader(header.encode('utf_8'), val.encode('utf_8'))
        return encoded.body.encode('utf-8')

    def render_GET(self, request):
        query = dict((k, v[0]) for k, v in request.args.iteritems())

        oid_request = self.oidserver.decodeRequest(query)

        if oid_request is None:
            return request.redirect_page('/')

        if oid_request.mode == 'checkid_setup':
            return self.start_auth(request, oid_request)
        if oid_request.mode == 'checkid_immediate':
            return self.oid_response(
                request, oid_request.answer(False))
        else:
            return self.oid_response(
                request,
                self.oidserver.handleRequest(oid_request))

    render_POST = render_GET

class AuthPage(Resource, object):
    def __init__(self, host, port, *args, **kwargs):
        super(AuthPage, self).__init__()
        self.host = host
        self.port = port
        self.token_page = TokenPage(*args, **kwargs)

    def render_GET(self, request):
        if 'user' in request.session:
            return request.redirect_page('/user?')
        return request.render_template('instructions.html',
                      host=self.host, port=self.port)

    def getChild(self, path, request):
        return self.token_page

class TokenPage(Resource):
    def __init__(self, oid):
        Resource.__init__(self)
        self.oid = oid

    def render_GET(self, request):
        token = request.prepath[-1].replace('-', '').lower()
        authname = log.callWithLogger(
            self.oid.storage,
            self.oid.storage.use_token, request.getClientIP(), token)
        if authname:
            request.session['user'] = authname
            request.push_session()
            return request.redirect_page('/user?')
        else:
            return request.render_template('wrong.html')

class IdPage(Resource):
    isLeaf = True

    def render_GET(self, request):
        gid = request.path.strip('/')
        return request.render_template('idpage.html', gid=gid)

class UserPage(Resource, object):
    isLeaf = True

    def __init__(self, storage, oid):
        super(UserPage, self).__init__()
        self.storage = storage
        self.oid = oid

    def render_GET(self, request):
        if 'user' not in request.session:
            return request.redirect_page('/auth')
        elif request.postpath:
            return request.redirect_page('/user?')
        else:
            return self.show_user_page(request)

    def render_POST(self, request):
        action, = request.postpath
        sess = request.session
        if not request.csrf.check():
            return request.redirect_page('/user?')
        elif 'user' not in sess:
            return request.redirect_page('/auth')
        elif action == 'complete':
            return self.complete(request)
        elif action == 'account':
            return self.account(request)
        else:
            return self.error(request, 'Unknown action.')

    def make_etag(self, request):
        return '"{0}"'.format(time())

    def show_user_page(self, request):
        authreqs = authreqs_from_session(request.session)
        request.setETag(self.make_etag(request))
        return request.render_template(
            'user.html',
            requests=authreqs
            )

    def error(self, request, message):
        request.toasts.error(message)
        return request.redirect_page('/user?')

    def complete(self, request):
        submit = None
        for key, vals in request.args.items():
            if key.startswith('submit-'):
                if submit:
                    return self.error(request, 'Multiple submit buttons used.')
                submit = key
        if submit is None:
            return self.error(request, 'No submit button used.')
        try:
            _, action, tr = submit.split('-', 2)
        except ValueError:
            return self.error(request, 'Bad format: {0}'.format(submit))
        authreqs = authreqs_from_session(request.session)
        authreq = authreqs.get(tr)
        if action == 'auth':
            return self.authenticate(request, authreq)
        elif action == 'cancel':
            return self.cancel(request, authreq)
        else:
            return self.error(request, 'Unknown action.')

    def authenticate(self, request, authreq):
        if not authreq:
            return self.error(request, 'Expired or invalid auth request.')
        sess = request.session
        sess['authreqs'].pop(authreq.trust_root, None)
        request.push_session()
        log.msg('authenticated {0} from {1} to {2}'.format(
            sess['user'], request.getClientIP(), authreq.trust_root))
        return self.oid.oid_response(request, authreq.answer(
            True, identity=oid_from_gid(request, sess['user'])))

    def cancel(self, request, authreq):
        if not authreq:
            return self.show_user_page(request)
        request.session['authreqs'].pop(authreq.trust_root, None)
        request.push_session()
        return self.oid.oid_response(request, authreq.answer(False))

    def account(self, request):
        if 'logout' in request.args:
            return self.logout(request)
        elif 'remember' in request.args:
            return self.remember(request)
        elif 'forget' in request.args:
            return self.remember(request, persist=False)
        else:
            return self.error(request, 'Unknown action.')

    def logout(self, request):
        sess = request.session
        sess.pop('user', None)
        sess.pop('csrf', None)
        request.toasts.info('Logged out succesfully.')
        request.push_session()
        return request.redirect_page('/auth')

    def remember(self, request, persist=True):
        sess = request.session
        sess['persist'] = persist
        if persist:
            request.toasts.info('You will stay logged in for at least '
                                '31 days.')
        else:
            request.toasts.info('You will be logged out at the end of this '
                                'browser session.')
        request.push_session()
        return request.redirect_page('/user?')

class Root(Resource):
    def __init__(self, idpage, storage):
        Resource.__init__(self)
        self.idpage = idpage
        self.storage = storage

    def render_GET(self, request):
        return request.render_template('frontpage.html')

    def getChild(self, path, request):
        if not path:
            return self
        if '@' in request.path:
            return self.idpage
        return Resource.getChild(self, path, request)

class FileNoListings(File):
    """Disable directory listings"""

    def directoryListing(self):
        return NoResource()

def make_site(oidstore, storage, arma_host, arma_port, address, serializer):
    oidserver = openidserver.Server(
        oidstore, 'http://{0}/openid'.format(address))
    session = SessionGetter(serializer)

    root = Root(IdPage(), storage)
    oid = OpenIdEndpoint(oidserver, storage)
    root.putChild('openid', oid)
    root.putChild('auth', AuthPage(arma_host, arma_port, oid))
    root.putChild('user', UserPage(storage, oid))
    root.putChild('static',
        FileNoListings(resource_filename(__name__, 'static')))

    site = Site(root)
    site.session = session
    site.requestFactory = NiceRequest
    return site

