import string
from time import time
from random import SystemRandom
from weakref import WeakKeyDictionary

from twisted.web.server import Request
import itsdangerous
import jinja2


def guess_autoescape(template_name):
    if template_name is None or '.' not in template_name:
        return False
    ext = template_name.rsplit('.', 1)[1]
    return ext in ('html', 'htm', 'xml')

templater = jinja2.Environment(
    autoescape=guess_autoescape,
    trim_blocks=True, lstrip_blocks=True,
    loader=jinja2.PackageLoader(__name__),
    extensions=['jinja2.ext.autoescape'],
    )

def seconds_until(ts):
    return ts - time()
templater.filters['seconds_until'] = seconds_until

def minutes_until(ts):
    return int(ts - time()) // 60
templater.filters['minutes_until'] = minutes_until

def minutes_until_str(ts):
    minutes = minutes_until(ts)
    if minutes < 2:
        return 'soon'
    else:
        return 'in {0} minutes'.format(minutes)
templater.filters['minutes_until_str'] = minutes_until_str


PERSIST_TIME = 60 * 60 * 24 * 31
class SessionGetter(object):
    def __init__(self, serializer, key='session'):
        self.instances = WeakKeyDictionary()
        self.serializer = serializer
        self.key = key

    def default(self, request):
        return {
            u'ip': request.getClientIP(),
            u'persist': False,
            }

    def _get(self, request):
        cookie = request.getCookie(self.key)
        if cookie:
            try:
                obj = self.serializer.loads(cookie)
            except itsdangerous.BadSignature:
                pass
            else:
                if ('ip' in obj and (obj['ip'] is None
                        or obj['ip'] == request.getClientIP())):
                    return obj
        return self.default(request)

    def get(self, request):
        if request in self.instances:
            return self.instances[request]
        else:
            ret = self.instances[request] = self._get(request)
            return ret

    def push(self, request):
        val = self.get(request)
        max_age = PERSIST_TIME if val[u'persist'] else None
        request.addCookie(
            self.key, self.serializer.dumps(val),
            path='/', max_age=max_age)

class SessionSaver(object):
    def __init__(self, request):
        self.request = request

    def __enter__(self):
        return self.request.session

    def __exit__(self, exc_typ, exc_val, exc_tb):
        self.request.push_session()

CSRF_ALPHABET = string.ascii_letters + string.digits
class CrossSiteRequestForgeryProtection(object):
    def __init__(self, request):
        self.request = request
        self.random = SystemRandom()

    def new_token(self):
        return ''.join(self.random.choice(CSRF_ALPHABET)
                       for i in range(32))

    def get_token(self):
        sess = self.request.session
        if 'csrf' in sess:
            return sess['csrf']
        else:
            ret = sess['csrf'] = self.new_token()
            self.request.push_session()
            return ret

    def html(self):
        return jinja2.Markup(
            '<input type="hidden" name="csrf" value="{0}" />'
            .format(self.get_token()))

    def check(self):
        sess = self.request.session
        if 'csrf' in sess and 'csrf' in self.request.args:
            if sess['csrf'] == self.request.args['csrf'][0]:
                return True
        return False

class MessageFlasher(object):
    def __init__(self, request):
        self.request = request

    def add(self, message, cls):
        sess = self.request.session
        if 'toasts' not in sess:
            toasts = sess['toasts'] = []
        else:
            toasts = sess['toasts']
        toasts.append((cls, message))
        self.request.push_session()

    def info(self, message):
        return self.add(message, 'info')

    def error(self, message):
        return self.add(message, 'error')

    def pop_messages(self):
        with SessionSaver(self.request) as sess:
            if 'toasts' in sess:
                for cls, message in sess['toasts'][:]:
                    sess['toasts'].pop(0)
                    yield cls, message

    def __iter__(self):
        sess = self.request.session
        if 'toasts' in sess:
            return iter(sess['toasts'])
        else:
            return iter(())

    def __nonzero__(self):
        sess = self.request.session
        if 'toasts' in sess:
            return bool(sess['toasts'])
        else:
            return False

class NiceRequest(Request, object):
    @property
    def session(self):
        return self.site.session.get(self)

    def push_session(self):
        return self.site.session.push(self)

    @property
    def csrf(self):
        return CrossSiteRequestForgeryProtection(self)

    @property
    def toasts(self):
        return MessageFlasher(self)

    def render_template(self, template, *args, **kwargs):
        return (
            templater
                .get_template(template)
                .render(*args, request=self, session=self.session,
                        hostname=self.getHeader('host'), **kwargs)
                .encode('utf-8')
            )

    def redirect_page(self, target):
        self.redirect(target)
        return '<a href="{0}">Redirect</a>'.format(target)

